/*  Tehnica RAII
    Resource Aquisition Is Initialization

	The Rule of  Three:
	 - Daca suntem obligati sa implementam una dintre functiile membre speciale:
	   1. constructor (inclusiv constructor de copiere)
	   2. destructor
	   3. operator de atribuire
	 - Atunci suntem obligati sa le implementam
	   PE TOATE TREI
*/

#include <iostream>
#include <vector>

template <class T>
void Print(const T& v)
{
	for (typename T::iterator it = v.begin(); it != v.end(); ++it)
		std::cout << *it << ' ';
	std::cout << '\n';
}


template <class T>
class Vector {
public:
    Vector() : n {0}, ptr {nullptr}
    {}

// ============= RAII =================================
    Vector(int n, const T& val = T()) : // achizitionarea resurselor la initializare
        n {n}, ptr {new T[n]}
    {
        for (int i = 0; i < n; ++i)
            ptr[i] = val;
  //      std::cout << __PRETTY_FUNCTION__ << '\n';
    }

	// Copy- Constructor (Deep Copy - Copie profunda)
	Vector(const Vector& v) : n {v.n}, ptr {new T[n]}
	{
		for (int i = 0; i < n; ++i)
			ptr[i] = v.ptr[i];
		std::cout << __PRETTY_FUNCTION__ << '\n';
	}
	
	Vector& operator = (const Vector& v)
	{
		if (&v == this)        // evitam auto-atribuirea
			return *this;
			
		if (ptr != nullptr)
			delete[] ptr;      // elibearea resurselor anterioare
		
		n = v.n;
		ptr = new T[v.n];      // realocare de resurse
		
		// copierea valorilor
		
		for (int i = 0; i < n; ++i)
			ptr[i] = v.ptr[i];
		
		std::cout << __PRETTY_FUNCTION__ << '\n';		
		return *this;
	}
	
    ~Vector()
    {
        if (ptr != nullptr)
            delete[] ptr;      // eliberarea resurselor
   //     std::cout << __PRETTY_FUNCTION__ << '\n';
    }
    
// ===============================================================


// Tip membru
    using iterator = T*;

// Functii Membre

    iterator begin() const
    {
        return ptr;
    }

    iterator end() const
    {
        return ptr + n;
    }

private:
    int n;
    T *ptr;
};

int main()
{
    Vector<int> v1(10, -7);
    
    {
		Vector<int> v2 = v1; // Default Copy-Constructor
    
    }

	Print(v1);
	
	{
		Vector<int> v2(5, 8);
	
		v1 = v2;      // ATIBUIRE
	}
	
	Print(v1);
	
}


